@@ -6,7 +6,7 @@ use Carp;
 use Perl::PrereqScanner::NotQuiteLite::Context;
 use Perl::PrereqScanner::NotQuiteLite::Util;
 
-our $VERSION = '0.49_02';
+our $VERSION = '0.49_03';
 
 our @BUNDLED_PARSERS = qw/
   Aliased Core Moose Plack POE Superclass
@@ -51,7 +51,7 @@ my %unsupported_packages = map {$_ => 1} qw(
 
 my %is_conditional = map {$_ => 1} qw(
   if elsif unless else given when
-  for foreach while until eval
+  for foreach while until
 );
 
 my %expects_expr_block = map {$_ => 1} qw(
@@ -238,9 +238,12 @@ sub scan_string {
 
   $string = '' unless defined $string;
 
+  my $c = Perl::PrereqScanner::NotQuiteLite::Context->new(%$self);
+
   # UTF8 BOM
   if ($string =~ s/\A(\xef\xbb\xbf)//s) {
     utf8::decode($string);
+    $c->{decoded} = 1;
   }
   # Other BOMs (TODO: also decode?)
   $string =~ s/\A(\x00\x00\xfe\xff|\xff\xfe\x00\x00|\xfe\xff|\xff\xfe)//s;
@@ -256,8 +259,6 @@ sub scan_string {
     $string =~ s/(?:\015\012|\015|\012)/\n/gs;
   }
 
-  my $c = Perl::PrereqScanner::NotQuiteLite::Context->new(%$self);
-
   # FIXME
   $c->{stack} = [];
   $c->{errors} = [];
@@ -270,9 +271,16 @@ sub scan_string {
 
   pos($string) = 0;
 
-  local $@;
-  eval { $self->_scan($c, \$string, 0) };
-  push @{$c->{errors}}, $@ if $@;
+  {
+    local $@;
+    eval { $self->_scan($c, \$string, 0) };
+    push @{$c->{errors}}, $@ if $@;
+    if ($c->{redo}) {
+      delete $c->{redo};
+      delete $c->{ended};
+      redo;
+    }
+  }
   $c;
 }
 
@@ -1229,8 +1237,8 @@ sub _scan {
               $c->{stack} = $saved_stack;
             }
           }
+          $current_scope &= MASK_EVAL;
         }
-        $current_scope &= MASK_EVAL;
         $c->{eval} = ($current_scope | $parent_scope) & F_EVAL ? 1 : 0;
       }
       if ($token eq 'eval') {
@@ -1828,10 +1836,13 @@ _debug("USE TOKENS: ".(Data::Dump::dump($tokens))) if !!DEBUG;
   if ($name eq 'utf8') {
     $c->add($name => 0);
     $c->{utf8} = 1;
-    _debug("UTF8 IS ON") if !!DEBUG;
-    my $pos = pos($$rstr); # perl 5.14/5.16 seem to reset pos by the following utf8::decode
-    utf8::decode($$rstr);
-    pos($$rstr) = $pos;
+    if (!$c->{decoded}) {
+      $c->{decoded} = 1;
+      _debug("UTF8 IS ON") if !!DEBUG;
+      utf8::decode($$rstr);
+      pos($$rstr) = 0;
+      $c->{ended} = $c->{redo} = 1;
+    }
   }
 
   if (is_module_name($name)) {
@@ -1867,13 +1878,13 @@ _debug("REQUIRE TOKENS: ".(Data::Dump::dump($tokens))) if !!DEBUG;
 
   my $c1 = substr($name, 0, 1);
   if ($c1 eq '5') {
-    $c->add(perl => $name);
+    $c->add_recommendation(perl => $name);
     return;
   }
   if ($c1 eq 'v') {
     my $c2 = substr($name, 1, 1);
     if ($c2 eq '5') {
-      $c->add(perl => $name);
+      $c->add_recommendation(perl => $name);
       return;
     }
     if ($c2 eq '6') {
@@ -1883,7 +1894,7 @@ _debug("REQUIRE TOKENS: ".(Data::Dump::dump($tokens))) if !!DEBUG;
     }
   }
   if ($name =~ /\A(\w|::)+\z/) {
-    $c->add($name => 0);
+    $c->add_recommendation($name => 0);
     return;
   }
 }
